"""Nvim API subpackage.

This package implements a higher-level API that wraps msgpack-rpc `Session`
instances.
"""

from pynvim.api.buffer import Buffer
from pynvim.api.common import decode_if_bytes, walk
from pynvim.api.nvim import Nvim, NvimError
from pynvim.api.tabpage import Tabpage
from pynvim.api.window import Window


__all__ = ('Nvim', 'Buffer', 'Window', 'Tabpage', 'NvimError',
           'decode_if_bytes', 'walk')



"""Shared utility functions."""

import sys
from traceback import format_exception
from typing import Any, Dict, Optional, Tuple, TypeVar

from pynvim._version import VERSION


def format_exc_skip(skip: int, limit: Optional[int] = None) -> str:
    """Like traceback.format_exc but allow skipping the first frames."""
    etype, val, tb = sys.exc_info()
    for _ in range(skip):
        if tb is not None:
            tb = tb.tb_next
    return ("".join(format_exception(etype, val, tb, limit))).rstrip()


T1 = TypeVar("T1")
T2 = TypeVar("T2")


def get_client_info(
    kind: str, type_: T1, method_spec: T2
) -> Tuple[str, Dict[str, Any], T1, T2, Dict[str, str]]:
    """Returns a tuple describing the client."""
    name = "python{}-{}".format(sys.version_info[0], kind)
    attributes = {"license": "Apache v2", "website": "github.com/neovim/pynvim"}
    return (name, VERSION.__dict__, type_, method_spec, attributes)



"""Asynchronous msgpack-rpc handling in the event loop pipeline."""
import logging
from traceback import format_exc
from typing import Any, AnyStr, Callable, Dict

from pynvim.msgpack_rpc.msgpack_stream import MsgpackStream

logger = logging.getLogger(__name__)
debug, info, warn = (logger.debug, logger.info, logger.warning,)


# response call back takes two arguments: (err, return_value)
ResponseCallback = Callable[..., None]


class AsyncSession:

    """Asynchronous msgpack-rpc layer that wraps a msgpack stream.

    This wraps the msgpack stream interface for reading/writing msgpack
    documents and exposes an interface for sending and receiving msgpack-rpc
    requests and notifications.
    """

    def __init__(self, msgpack_stream: MsgpackStream):
        """Wrap `msgpack_stream` on a msgpack-rpc interface."""
        self._msgpack_stream = msgpack_stream
        self._next_request_id = 1
        self._pending_requests: Dict[int, ResponseCallback] = {}
        self._request_cb = self._notification_cb = None
        self._handlers = {
            0: self._on_request,
            1: self._on_response,
            2: self._on_notification
        }
        self.loop = msgpack_stream.loop

    def threadsafe_call(self, fn):
        """Wrapper around `MsgpackStream.threadsafe_call`."""
        self._msgpack_stream.threadsafe_call(fn)

    def request(self, method: AnyStr, args: Any,
                response_cb: ResponseCallback) -> None:
        """Send a msgpack-rpc request to Nvim.

        A msgpack-rpc with method `method` and argument `args` is sent to
        Nvim. The `response_cb` function is called with when the response
        is available.
        """
        request_id = self._next_request_id
        self._next_request_id = request_id + 1
        self._msgpack_stream.send([0, request_id, method, args])
        self._pending_requests[request_id] = response_cb

    def notify(self, method, args):
        """Send a msgpack-rpc notification to Nvim.

        A msgpack-rpc with method `method` and argument `args` is sent to
        Nvim. This will have the same effect as a request, but no response
        will be received
        """
        self._msgpack_stream.send([2, method, args])

    def run(self, request_cb, notification_cb):
        """Run the event loop to receive requests and notifications from Nvim.

        While the event loop is running, `request_cb` and `notification_cb`
        will be called whenever requests or notifications are respectively
        available.
        """
        self._request_cb = request_cb
        self._notification_cb = notification_cb
        self._msgpack_stream.run(self._on_message)
        self._request_cb = None
        self._notification_cb = None

    def stop(self):
        """Stop the event loop."""
        self._msgpack_stream.stop()

    def close(self):
        """Close the event loop."""
        self._msgpack_stream.close()

    def _on_message(self, msg):
        try:
            self._handlers.get(msg[0], self._on_invalid_message)(msg)
        except Exception:
            err_str = format_exc(5)
            pass # replaces next logging statement
            # warn(err_str)
            self._msgpack_stream.send([1, 0, err_str, None])

    def _on_request(self, msg):
        # request
        #   - msg[1]: id
        #   - msg[2]: method name
        #   - msg[3]: arguments
        pass # replaces next logging statement
        # debug('received request: %s, %s', msg[2], msg[3])
        assert self._request_cb is not None
        self._request_cb(msg[2], msg[3],
                         Response(self._msgpack_stream, msg[1]))

    def _on_response(self, msg):
        # response to a previous request:
        #   - msg[1]: the id
        #   - msg[2]: error(if any)
        #   - msg[3]: result(if not errored)
        pass # replaces next logging statement
        # debug('received response: %s, %s', msg[2], msg[3])
        self._pending_requests.pop(msg[1])(msg[2], msg[3])

    def _on_notification(self, msg):
        # notification/event
        #   - msg[1]: event name
        #   - msg[2]: arguments
        pass # replaces next logging statement
        # debug('received notification: %s, %s', msg[1], msg[2])
        assert self._notification_cb is not None
        self._notification_cb(msg[1], msg[2])

    def _on_invalid_message(self, msg):
        error = 'Received invalid message %s' % msg
        pass # replaces next logging statement
        # warn(error)
        self._msgpack_stream.send([1, 0, error, None])


class Response:
    """Response to a msgpack-rpc request that came from Nvim.

    When Nvim sends a msgpack-rpc request, an instance of this class is
    created for remembering state required to send a response.
    """

    def __init__(self, msgpack_stream: MsgpackStream, request_id: int):
        """Initialize the Response instance."""
        self._msgpack_stream = msgpack_stream
        self._request_id = request_id

    def send(self, value, error=False):
        """Send the response.

        If `error` is True, it will be sent as an error.
        """
        if error:
            resp = [1, self._request_id, value, None]
        else:
            resp = [1, self._request_id, None, value]
        pass # replaces next logging statement
        # debug('sending response to request %d: %s', self._request_id, resp)
        self._msgpack_stream.send(resp)



# type: ignore
# pylint: disable=protected-access
import os
from typing import Sequence

from pynvim.plugin.host import Host, host_method_spec
from pynvim.plugin.script_host import ScriptHost

__PATH__ = os.path.abspath(os.path.dirname(__file__))


def test_host_imports(vim):
    h = ScriptHost(vim)
    try:
        assert h.module.__dict__['vim']
        assert h.module.__dict__['vim'] == h.legacy_vim
        assert h.module.__dict__['sys']
    finally:
        h.teardown()


def test_host_import_rplugin_modules(vim):
    # Test whether a Host can load and import rplugins (#461).
    # See also $VIMRUNTIME/autoload/provider/pythonx.vim.
    h = Host(vim)

    plugins: Sequence[str] = [  # plugin paths like real rplugins
        os.path.join(__PATH__, "./fixtures/simple_plugin/rplugin/python3/simple_nvim.py"),
        os.path.join(__PATH__, "./fixtures/module_plugin/rplugin/python3/mymodule/"),
        os.path.join(__PATH__, "./fixtures/module_plugin/rplugin/python3/mymodule"),  # duplicate
    ]
    h._load(plugins)
    assert len(h._loaded) == 2

    # pylint: disable-next=unbalanced-tuple-unpacking
    simple_nvim, mymodule = list(h._loaded.values())
    assert simple_nvim['module'].__name__ == 'simple_nvim'
    assert mymodule['module'].__name__ == 'mymodule'


def test_host_clientinfo(vim):
    h = Host(vim)
    assert h._request_handlers.keys() == host_method_spec.keys()
    assert 'remote' == vim.api.get_chan_info(vim.channel_id)['client']['type']
    h._load([])
    assert 'host' == vim.api.get_chan_info(vim.channel_id)['client']['type']


# Smoke test for Host._on_error_event(). #425
def test_host_async_error(vim):
    h = Host(vim)
    h._load([])
    # Invoke a bogus Ex command via notify (async).
    vim.command("lolwut", async_=True)
    event = vim.next_message()
    assert event[1] == 'nvim_error_event'
    assert 'rplugin-host: Async request caused an error:\nboom\n' \
           in h._on_error_event(None, 'boom')



import os

import pytest

from pynvim.api import Nvim, NvimError


def test_repr(vim: Nvim) -> None:
    assert repr(vim.current.buffer) == "<Buffer(handle=1)>"


def test_get_length(vim: Nvim) -> None:
    assert len(vim.current.buffer) == 1
    vim.current.buffer.append('line')
    assert len(vim.current.buffer) == 2
    vim.current.buffer.append('line')
    assert len(vim.current.buffer) == 3
    vim.current.buffer[-1] = None
    assert len(vim.current.buffer) == 2
    vim.current.buffer[-1] = None
    vim.current.buffer[-1] = None
    # There's always at least one line
    assert len(vim.current.buffer) == 1


def test_get_set_del_line(vim: Nvim) -> None:
    assert vim.current.buffer[0] == ''
    vim.current.buffer[0] = 'line1'
    assert vim.current.buffer[0] == 'line1'
    vim.current.buffer[0] = 'line2'
    assert vim.current.buffer[0] == 'line2'
    vim.current.buffer[0] = None
    assert vim.current.buffer[0] == ''
    # __delitem__
    vim.current.buffer[:] = ['line1', 'line2', 'line3']
    assert vim.current.buffer[2] == 'line3'
    del vim.current.buffer[0]
    assert vim.current.buffer[0] == 'line2'
    assert vim.current.buffer[1] == 'line3'
    del vim.current.buffer[-1]
    assert vim.current.buffer[0] == 'line2'
    assert len(vim.current.buffer) == 1


def test_get_set_del_slice(vim: Nvim) -> None:
    assert vim.current.buffer[:] == ['']
    # Replace buffer
    vim.current.buffer[:] = ['a', 'b', 'c']
    assert vim.current.buffer[:] == ['a', 'b', 'c']
    assert vim.current.buffer[1:] == ['b', 'c']
    assert vim.current.buffer[1:2] == ['b']
    assert vim.current.buffer[1:1] == []
    assert vim.current.buffer[:-1] == ['a', 'b']
    assert vim.current.buffer[1:-1] == ['b']
    assert vim.current.buffer[-2:] == ['b', 'c']
    vim.current.buffer[1:2] = ['a', 'b', 'c']
    assert vim.current.buffer[:] == ['a', 'a', 'b', 'c', 'c']
    vim.current.buffer[-1:] = ['a', 'b', 'c']
    assert vim.current.buffer[:] == ['a', 'a', 'b', 'c', 'a', 'b', 'c']
    vim.current.buffer[:-3] = None
    assert vim.current.buffer[:] == ['a', 'b', 'c']
    vim.current.buffer[:] = None
    assert vim.current.buffer[:] == ['']
    # __delitem__
    vim.current.buffer[:] = ['a', 'b', 'c']
    del vim.current.buffer[:]
    assert vim.current.buffer[:] == ['']
    vim.current.buffer[:] = ['a', 'b', 'c']
    del vim.current.buffer[:1]
    assert vim.current.buffer[:] == ['b', 'c']
    del vim.current.buffer[:-1]
    assert vim.current.buffer[:] == ['c']


def test_vars(vim: Nvim) -> None:
    vim.current.buffer.vars['python'] = [1, 2, {'3': 1}]
    assert vim.current.buffer.vars['python'] == [1, 2, {'3': 1}]
    assert vim.eval('b:python') == [1, 2, {'3': 1}]
    assert vim.current.buffer.vars.get('python') == [1, 2, {'3': 1}]

    del vim.current.buffer.vars['python']
    with pytest.raises(KeyError):
        vim.current.buffer.vars['python']
    assert vim.eval('exists("b:python")') == 0

    with pytest.raises(KeyError):
        del vim.current.buffer.vars['python']

    assert vim.current.buffer.vars.get('python', 'default') == 'default'


def test_api(vim: Nvim) -> None:
    vim.current.buffer.api.set_var('myvar', 'thetext')
    assert vim.current.buffer.api.get_var('myvar') == 'thetext'
    assert vim.eval('b:myvar') == 'thetext'
    vim.current.buffer.api.set_lines(0, -1, True, ['alpha', 'beta'])
    assert vim.current.buffer.api.get_lines(0, -1, True) == ['alpha', 'beta']
    assert vim.current.buffer[:] == ['alpha', 'beta']


def test_options(vim: Nvim) -> None:
    assert vim.current.buffer.options['shiftwidth'] == 8
    vim.current.buffer.options['shiftwidth'] = 4
    assert vim.current.buffer.options['shiftwidth'] == 4
    # global-local option
    global_define = vim.options['define']
    vim.current.buffer.options['define'] = 'test'
    assert vim.current.buffer.options['define'] == 'test'
    # Doesn't change the global value
    assert vim.options['define'] == global_define

    with pytest.raises(KeyError) as excinfo:
        vim.current.buffer.options['doesnotexist']
    assert excinfo.value.args == ("Invalid option name: 'doesnotexist'",)


def test_number(vim: Nvim) -> None:
    curnum = vim.current.buffer.number
    vim.command('new')
    assert vim.current.buffer.number == curnum + 1
    vim.command('new')
    assert vim.current.buffer.number == curnum + 2


def test_name(vim: Nvim) -> None:
    vim.command('new')
    assert vim.current.buffer.name == ''
    new_name = vim.eval('resolve(tempname())')
    vim.current.buffer.name = new_name
    assert vim.current.buffer.name == new_name
    vim.command('silent w!')
    assert os.path.isfile(new_name)
    os.unlink(new_name)


def test_valid(vim: Nvim) -> None:
    vim.command('new')
    buffer = vim.current.buffer
    assert buffer.valid
    vim.command('bw!')
    assert not buffer.valid


def test_append(vim: Nvim) -> None:
    vim.current.buffer.append('a')
    assert vim.current.buffer[:] == ['', 'a']
    vim.current.buffer.append('b', 0)
    assert vim.current.buffer[:] == ['b', '', 'a']
    vim.current.buffer.append(['c', 'd'])
    assert vim.current.buffer[:] == ['b', '', 'a', 'c', 'd']
    vim.current.buffer.append(['c', 'd'], 2)
    assert vim.current.buffer[:] == ['b', '', 'c', 'd', 'a', 'c', 'd']
    vim.current.buffer.append(b'bytes')
    assert vim.current.buffer[:] == ['b', '', 'c', 'd', 'a', 'c', 'd', 'bytes']


def test_mark(vim: Nvim) -> None:
    vim.current.buffer.append(['a', 'bit of', 'text'])
    vim.current.window.cursor = (3, 4)
    vim.command('mark V')
    assert vim.current.buffer.mark('V') == (3, 0)


def test_invalid_utf8(vim: Nvim) -> None:
    vim.command('normal "=printf("%c", 0xFF)\np')
    assert vim.eval("char2nr(getline(1))") == 0xFF
    assert vim.current.buffer[:] == ['\udcff']

    vim.current.line += 'x'
    assert vim.eval("getline(1)", decode=False) == '\udcffx'
    assert vim.current.buffer[:] == ['\udcffx']


def test_get_exceptions(vim: Nvim) -> None:
    with pytest.raises(KeyError) as excinfo:
        vim.current.buffer.options['invalid-option']

    assert not isinstance(excinfo.value, NvimError)
    assert excinfo.value.args == ("Invalid option name: 'invalid-option'",)


def test_set_items_for_range(vim: Nvim) -> None:
    vim.current.buffer[:] = ['a', 'b', 'c', 'd', 'e']
    r = vim.current.buffer.range(1, 3)
    r[1:3] = ['foo'] * 3
    assert vim.current.buffer[:] == ['a', 'foo', 'foo', 'foo', 'd', 'e']


# NB: we can't easily test the effect of this. But at least run the lua
# function sync, so we know it runs without runtime error with simple args.
def test_update_highlights(vim: Nvim) -> None:
    vim.current.buffer[:] = ['a', 'b', 'c']
    src_id = vim.new_highlight_source()
    vim.current.buffer.update_highlights(
        src_id, [("Comment", 0, 0, -1), ("String", 1, 0, 1)], clear=True, async_=False
    )


def test_buffer_inequality(vim: Nvim) -> None:
    b = vim.current.buffer
    assert not (b != vim.current.buffer)



"""Main Nvim interface."""

from __future__ import annotations

import asyncio
import os
import sys
import threading
from functools import partial
from traceback import format_stack
from types import SimpleNamespace
from typing import (Any, AnyStr, Callable, Dict, Iterator, List, Optional,
                    TYPE_CHECKING, Union)

from msgpack import ExtType

from pynvim.api.buffer import Buffer
from pynvim.api.common import (NvimError, Remote, RemoteApi, RemoteMap, RemoteSequence,
                               TDecodeMode, decode_if_bytes, walk)
from pynvim.api.tabpage import Tabpage
from pynvim.api.window import Window
from pynvim.util import format_exc_skip

if TYPE_CHECKING:
    from pynvim.msgpack_rpc import Session

if sys.version_info < (3, 8):
    from typing_extensions import Literal
else:
    from typing import Literal

__all__ = ['Nvim']


os_chdir = os.chdir

lua_module = """
local a = vim.api
local function update_highlights(buf, src_id, hls, clear_first, clear_end)
  if clear_first ~= nil then
      a.nvim_buf_clear_highlight(buf, src_id, clear_first, clear_end)
  end
  for _,hl in pairs(hls) do
    local group, line, col_start, col_end = unpack(hl)
    if col_start == nil then
      col_start = 0
    end
    if col_end == nil then
      col_end = -1
    end
    a.nvim_buf_add_highlight(buf, src_id, group, line, col_start, col_end)
  end
end

local chid = ...
local mod = {update_highlights=update_highlights}
_G["_pynvim_"..chid] = mod
"""


class Nvim:
    """Class that represents a remote Nvim instance.

    This class is main entry point to Nvim remote API, it is a wrapper
    around Session instances.

    The constructor of this class must not be called directly. Instead, the
    `from_session` class method should be used to create the first instance
    from a raw `Session` instance.

    Subsequent instances for the same session can be created by calling the
    `with_decode` instance method to change the decoding behavior or
    `SubClass.from_nvim(nvim)` where `SubClass` is a subclass of `Nvim`, which
    is useful for having multiple `Nvim` objects that behave differently
    without one affecting the other.

    When this library is used on python3.4+, asyncio event loop is guaranteed
    to be used. It is available as the "loop" attribute of this class. Note
    that asyncio callbacks cannot make blocking requests, which includes
    accessing state-dependent attributes. They should instead schedule another
    callback using nvim.async_call, which will not have this restriction.
    """

    @classmethod
    def from_session(cls, session: Session) -> Nvim:
        """Create a new Nvim instance for a Session instance.

        This method must be called to create the first Nvim instance, since it
        queries Nvim metadata for type information and sets a SessionHook for
        creating specialized objects from Nvim remote handles.
        """
        session.error_wrapper = lambda e: NvimError(decode_if_bytes(e[1]))
        channel_id, metadata = session.request(b'nvim_get_api_info')

        metadata = walk(decode_if_bytes, metadata)

        types = {
            metadata['types']['Buffer']['id']: Buffer,
            metadata['types']['Window']['id']: Window,
            metadata['types']['Tabpage']['id']: Tabpage,
        }

        return cls(session, channel_id, metadata, types)

    @classmethod
    def from_nvim(cls, nvim: Nvim) -> Nvim:
        """Create a new Nvim instance from an existing instance."""
        return cls(nvim._session, nvim.channel_id, nvim.metadata,
                   nvim.types, nvim._decode, nvim._err_cb)

    def __init__(
        self,
        session: Session,
        channel_id: int,
        metadata: Dict[str, Any],
        types: Dict[int, Any],
        decode: TDecodeMode = True,
        err_cb: Optional[Callable[[str], None]] = None
    ):
        """Initialize a new Nvim instance. This method is module-private."""
        self._session = session
        self.channel_id = channel_id
        self.metadata = metadata
        version = metadata.get("version", {"api_level": 0})
        self.version = SimpleNamespace(**version)
        self.types = types
        self.api = RemoteApi(self, 'nvim_')
        self.vars = RemoteMap(self, 'nvim_get_var', 'nvim_set_var', 'nvim_del_var')
        self.vvars = RemoteMap(self, 'nvim_get_vvar', None, None)
        self.options = RemoteMap(self, 'nvim_get_option', 'nvim_set_option')
        self.buffers = Buffers(self)
        self.windows: RemoteSequence[Window] = RemoteSequence(self, 'nvim_list_wins')
        self.tabpages: RemoteSequence[Tabpage] = RemoteSequence(
            self, 'nvim_list_tabpages'
        )
        self.current = Current(self)
        self.session = CompatibilitySession(self)
        self.funcs = Funcs(self)
        self.lua = LuaFuncs(self)
        self.error = NvimError
        self._decode = decode
        if err_cb is None:
            self._err_cb: Callable[[str], Any] = lambda _: None
        else:
            self._err_cb = err_cb

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        """Get the event loop (exposed to rplugins)."""  # noqa

        # see #294: for python 3.4+, the only available and guaranteed
        # implementation of msgpack_rpc BaseEventLoop is the AsyncioEventLoop.
        # The underlying asyncio event loop is exposed to rplugins.
        # pylint: disable=protected-access
        return self._session.loop._loop  # type: ignore

    def _from_nvim(self, obj: Any, decode: Optional[TDecodeMode] = None) -> Any:
        if decode is None:
            decode = self._decode
        if type(obj) is ExtType:
            cls = self.types[obj.code]
            return cls(self, (obj.code, obj.data))
        if decode:
            obj = decode_if_bytes(obj, decode)
        return obj

    def _to_nvim(self, obj: Any) -> Any:
        if isinstance(obj, Remote):
            return ExtType(*obj.code_data)
        return obj

    def _get_lua_private(self) -> LuaFuncs:
        if not getattr(self._session, "_has_lua", False):
            self.exec_lua(lua_module, self.channel_id)
            self._session._has_lua = True  # type: ignore[attr-defined]
        return getattr(self.lua, "_pynvim_{}".format(self.channel_id))

    def request(self, name: str, *args: Any, **kwargs: Any) -> Any:
        r"""Send an API request or notification to nvim.

        It is rarely needed to call this function directly, as most API
        functions have python wrapper functions. The `api` object can
        be also be used to call API functions as methods:

            vim.api.err_write('ERROR\n', async_=True)
            vim.current.buffer.api.get_mark('.')

        is equivalent to

            vim.request('nvim_err_write', 'ERROR\n', async_=True)
            vim.request('nvim_buf_get_mark', vim.current.buffer, '.')


        Normally a blocking request will be sent.  If the `async_` flag is
        present and True, a asynchronous notification is sent instead. This
        will never block, and the return value or error is ignored.
        """
        if (self._session._loop_thread is not None
                and threading.current_thread() != self._session._loop_thread):

            msg = ("Request from non-main thread.\n"
                   "Requests from different threads should be wrapped "
                   "with nvim.async_call(cb, ...) \n{}\n"
                   .format('\n'.join(format_stack(None, 5)[:-1])))

            self.async_call(self._err_cb, msg)
            raise NvimError("request from non-main thread")

        decode = kwargs.pop('decode', self._decode)
        args = walk(self._to_nvim, args)
        res = self._session.request(name, *args, **kwargs)
        return walk(partial(self._from_nvim, decode=decode), res)

    def next_message(self) -> Any:
        """Block until a message(request or notification) is available.

        If any messages were previously enqueued, return the first in queue.
        If not, run the event loop until one is received.
        """
        msg = self._session.next_message()
        if msg:
            return walk(self._from_nvim, msg)

    def run_loop(
        self,
        request_cb: Optional[Callable[[str, List[Any]], Any]],
        notification_cb: Optional[Callable[[str, List[Any]], Any]],
        setup_cb: Optional[Callable[[], None]] = None,
        err_cb: Optional[Callable[[str], Any]] = None
    ) -> None:
        """Run the event loop to receive requests and notifications from Nvim.

        This should not be called from a plugin running in the host, which
        already runs the loop and dispatches events to plugins.
        """
        if err_cb is None:
            err_cb = sys.stderr.write
        self._err_cb = err_cb

        def filter_request_cb(name: str, args: Any) -> Any:
            name = self._from_nvim(name)
            args = walk(self._from_nvim, args)
            try:
                result = request_cb(name, args)  # type: ignore[misc]
            except Exception:
                msg = ("error caught in request handler '{} {}'\n{}\n\n"
                       .format(name, args, format_exc_skip(1)))
                self._err_cb(msg)
                raise
            return walk(self._to_nvim, result)

        def filter_notification_cb(name: str, args: Any) -> None:
            name = self._from_nvim(name)
            args = walk(self._from_nvim, args)
            try:
                notification_cb(name, args)  # type: ignore[misc]
            except Exception:
                msg = ("error caught in notification handler '{} {}'\n{}\n\n"
                       .format(name, args, format_exc_skip(1)))
                self._err_cb(msg)
                raise

        self._session.run(filter_request_cb, filter_notification_cb, setup_cb)

    def stop_loop(self) -> None:
        """Stop the event loop being started with `run_loop`."""
        self._session.stop()

    def close(self) -> None:
        """Close the nvim session and release its resources."""
        self._session.close()

    def __enter__(self) -> Nvim:
        """Enter nvim session as a context manager."""
        return self

    def __exit__(self, *exc_info: Any) -> None:
        """Exit nvim session as a context manager.

        Closes the event loop.
        """
        self.close()

    def with_decode(self, decode: Literal[True] = True) -> Nvim:
        """Initialize a new Nvim instance."""
        return Nvim(self._session, self.channel_id,
                    self.metadata, self.types, decode, self._err_cb)

    def ui_attach(
        self, width: int, height: int, rgb: Optional[bool] = None, **kwargs: Any
    ) -> None:
        """Register as a remote UI.

        After this method is called, the client will receive redraw
        notifications.
        """
        options = kwargs
        if rgb is not None:
            options['rgb'] = rgb
        return self.request('nvim_ui_attach', width, height, options)

    def ui_detach(self) -> None:
        """Unregister as a remote UI."""
        return self.request('nvim_ui_detach')

    def ui_try_resize(self, width: int, height: int) -> None:
        """Notify nvim that the client window has resized.

        If possible, nvim will send a redraw request to resize.
        """
        return self.request('ui_try_resize', width, height)

    def subscribe(self, event: str) -> None:
        """Subscribe to a Nvim event."""
        return self.request('nvim_subscribe', event)

    def unsubscribe(self, event: str) -> None:
        """Unsubscribe to a Nvim event."""
        return self.request('nvim_unsubscribe', event)

    def command(self, string: str, **kwargs: Any) -> None:
        """Execute a single ex command."""
        return self.request('nvim_command', string, **kwargs)

    def command_output(self, string: str) -> str:
        """Execute a single ex command and return the output."""
        return self.request('nvim_command_output', string)

    def eval(self, string: str, **kwargs: Any) -> Any:
        """Evaluate a vimscript expression."""
        return self.request('nvim_eval', string, **kwargs)

    def call(self, name: str, *args: Any, **kwargs: Any) -> Any:
        """Call a vimscript function."""
        return self.request('nvim_call_function', name, args, **kwargs)

    def exec_lua(self, code: str, *args: Any, **kwargs: Any) -> Any:
        """Execute lua code.

        Additional parameters are available as `...` inside the lua chunk.
        Only statements are executed.  To evaluate an expression, prefix it
        with `return`: `return my_function(...)`

        There is a shorthand syntax to call lua functions with arguments:

            nvim.lua.func(1,2)
            nvim.lua.mymod.myfunction(data, async_=True)

        is equivalent to

            nvim.exec_lua("return func(...)", 1, 2)
            nvim.exec_lua("mymod.myfunction(...)", data, async_=True)

        Note that with `async_=True` there is no return value.
        """
        return self.request('nvim_execute_lua', code, args, **kwargs)

    def strwidth(self, string: str) -> int:
        """Return the number of display cells `string` occupies.

        Tab is counted as one cell.
        """
        return self.request('nvim_strwidth', string)

    def list_runtime_paths(self) -> List[str]:
        """Return a list of paths contained in the 'runtimepath' option."""
        return self.request('nvim_list_runtime_paths')

    def foreach_rtp(self, cb: Callable[[str], Any]) -> None:
        """Invoke `cb` for each path in 'runtimepath'.

        Call the given callable for each path in 'runtimepath' until either
        callable returns something but None, the exception is raised or there
        are no longer paths. If stopped in case callable returned non-None,
        vim.foreach_rtp function returns the value returned by callable.
        """
        for path in self.list_runtime_paths():
            try:
                if cb(path) is not None:
                    break
            except Exception:
                break

    def chdir(self, dir_path: str) -> None:
        """Run os.chdir, then all appropriate vim stuff."""
        os_chdir(dir_path)
        return self.request('nvim_set_current_dir', dir_path)

    def feedkeys(self, keys: str, options: str = '', escape_csi: bool = True) -> None:
        """Push `keys` to Nvim user input buffer."""
        return self.request('nvim_feedkeys', keys, options, escape_csi)

    def input(self, bytes: AnyStr) -> int:
        """Push `bytes` to Nvim low level input buffer.

        Unlike `feedkeys()`, this uses the lowest level input buffer and the
        call is not deferred. It returns the number of bytes actually
        written(which can be less than what was requested if the buffer is
        full).
        """
        return self.request('nvim_input', bytes)

    def replace_termcodes(
        self,
        string: str,
        from_part: bool = False,
        do_lt: bool = True,
        special: bool = True
    ) -> str:
        r"""Replace any terminal code strings by byte sequences.

        The returned sequences are Nvim's internal representation of keys,
        for example:

        <esc> -> '\x1b'
        <cr>  -> '\r'
        <c-l> -> '\x0c'
        <up>  -> '\x80ku'

        The returned sequences can be used as input to `feedkeys`.
        """
        return self.request('nvim_replace_termcodes', string,
                            from_part, do_lt, special)

    def out_write(self, msg: str, **kwargs: Any) -> None:
        r"""Print `msg` as a normal message.

        The message is buffered (won't display) until linefeed ("\n").
        """
        return self.request('nvim_out_write', msg, **kwargs)

    def err_write(self, msg: str, **kwargs: Any) -> None:
        r"""Print `msg` as an error message.

        The message is buffered (won't display) until linefeed ("\n").
        """
        if self._thread_invalid():
            # special case: if a non-main thread writes to stderr
            # i.e. due to an uncaught exception, pass it through
            # without raising an additional exception.
            self.async_call(self.err_write, msg, **kwargs)
            return
        return self.request('nvim_err_write', msg, **kwargs)

    def _thread_invalid(self) -> bool:
        return (self._session._loop_thread is not None
                and threading.current_thread() != self._session._loop_thread)

    def quit(self, quit_command: str = 'qa!') -> None:
        """Send a quit command to Nvim.

        By default, the quit command is 'qa!' which will make Nvim quit without
        saving anything.
        """
        try:
            self.command(quit_command)
        except OSError:
            # sending a quit command will raise an IOError because the
            # connection is closed before a response is received. Safe to
            # ignore it.
            pass

    def new_highlight_source(self) -> int:
        """Return new src_id for use with Buffer.add_highlight."""
        return self.current.buffer.add_highlight("", 0, src_id=0)

    def async_call(self, fn: Callable[..., Any], *args: Any, **kwargs: Any) -> None:
        """Schedule `fn` to be called by the event loop soon.

        This function is thread-safe, and is the only way code not
        on the main thread could interact with nvim api objects.

        This function can also be called in a synchronous
        event handler, just before it returns, to defer execution
        that shouldn't block neovim.
        """
        call_point = ''.join(format_stack(None, 5)[:-1])

        def handler() -> None:
            try:
                fn(*args, **kwargs)
            except Exception as err:
                msg = ("error caught while executing async callback:\n"
                       "{!r}\n{}\n \nthe call was requested at\n{}"
                       .format(err, format_exc_skip(1), call_point))
                self._err_cb(msg)
                raise
        self._session.threadsafe_call(handler)


class Buffers(object):

    """Remote NVim buffers.

    Currently the interface for interacting with remote NVim buffers is the
    `nvim_list_bufs` msgpack-rpc function. Most methods fetch the list of
    buffers from NVim.

    Conforms to *python-buffers*.
    """

    def __init__(self, nvim: Nvim):
        """Initialize a Buffers object with Nvim object `nvim`."""
        self._fetch_buffers = nvim.api.list_bufs

    def __len__(self) -> int:
        """Return the count of buffers."""
        return len(self._fetch_buffers())

    def __getitem__(self, number: int) -> Buffer:
        """Return the Buffer object matching buffer number `number`."""
        for b in self._fetch_buffers():
            if b.number == number:
                return b
        raise KeyError(number)

    def __contains__(self, b: Buffer) -> bool:
        """Return whether Buffer `b` is a known valid buffer."""
        return isinstance(b, Buffer) and b.valid

    def __iter__(self) -> Iterator[Buffer]:
        """Return an iterator over the list of buffers."""
        return iter(self._fetch_buffers())


class CompatibilitySession(object):

    """Helper class for API compatibility."""

    def __init__(self, nvim: Nvim):
        self.threadsafe_call = nvim.async_call


class Current(object):

    """Helper class for emulating vim.current from python-vim."""

    def __init__(self, session: Nvim):
        self._session = session
        self.range = None

    @property
    def line(self) -> str:
        return self._session.request('nvim_get_current_line')

    @line.setter
    def line(self, line: str) -> None:
        return self._session.request('nvim_set_current_line', line)

    @line.deleter
    def line(self) -> None:
        return self._session.request('nvim_del_current_line')

    @property
    def buffer(self) -> Buffer:
        return self._session.request('nvim_get_current_buf')

    @buffer.setter
    def buffer(self, buffer: Union[Buffer, int]) -> None:
        return self._session.request('nvim_set_current_buf', buffer)

    @property
    def window(self) -> Window:
        return self._session.request('nvim_get_current_win')

    @window.setter
    def window(self, window: Union[Window, int]) -> None:
        return self._session.request('nvim_set_current_win', window)

    @property
    def tabpage(self) -> Tabpage:
        return self._session.request('nvim_get_current_tabpage')

    @tabpage.setter
    def tabpage(self, tabpage: Union[Tabpage, int]) -> None:
        return self._session.request('nvim_set_current_tabpage', tabpage)


class Funcs:
    """Helper class for functional vimscript interface."""

    def __init__(self, nvim: Nvim):
        self._nvim = nvim

    def __getattr__(self, name: str) -> Callable[..., Any]:
        return partial(self._nvim.call, name)


class LuaFuncs:
    """Wrapper to allow lua functions to be called like python methods."""

    def __init__(self, nvim: Nvim, name: str = ""):
        self._nvim = nvim
        self.name = name

    def __getattr__(self, name: str) -> LuaFuncs:
        """Return wrapper to named api method."""
        prefix = self.name + "." if self.name else ""
        return LuaFuncs(self._nvim, prefix + name)

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        # first new function after keyword rename, be a bit noisy
        if 'async' in kwargs:
            raise ValueError('"async" argument is not allowed. '
                             'Use "async_" instead.')
        async_ = kwargs.get('async_', False)
        pattern = "return {}(...)" if not async_ else "{}(...)"
        code = pattern.format(self.name)
        return self._nvim.exec_lua(code, *args, **kwargs)



"""Nvim plugin/host subpackage."""

from pynvim.plugin.decorators import (autocmd, command, decode, encoding, function,
                                      plugin, rpc_export, shutdown_hook)
from pynvim.plugin.host import Host  # type: ignore[attr-defined]


__all__ = ('Host', 'plugin', 'rpc_export', 'command', 'autocmd',
           'function', 'encoding', 'decode', 'shutdown_hook')



import pytest

from pynvim.api import Nvim


def test_buffer(vim: Nvim) -> None:
    assert vim.current.buffer == vim.windows[0].buffer
    vim.command('new')
    vim.current.window = vim.windows[1]
    assert vim.current.buffer == vim.windows[1].buffer
    assert vim.windows[0].buffer != vim.windows[1].buffer


def test_cursor(vim: Nvim) -> None:
    assert vim.current.window.cursor == (1, 0)
    vim.command('normal ityping\033o  some text')
    assert vim.current.buffer[:] == ['typing', '  some text']
    assert vim.current.window.cursor == (2, 10)
    vim.current.window.cursor = (2, 6)
    vim.command('normal i dumb')
    assert vim.current.buffer[:] == ['typing', '  some dumb text']


def test_height(vim: Nvim) -> None:
    vim.command('vsplit')
    assert vim.windows[1].height == vim.windows[0].height
    vim.current.window = vim.windows[1]
    vim.command('split')
    assert vim.windows[1].height == vim.windows[0].height // 2
    vim.windows[1].height = 2
    assert vim.windows[1].height == 2


def test_width(vim: Nvim) -> None:
    vim.command('split')
    assert vim.windows[1].width == vim.windows[0].width
    vim.current.window = vim.windows[1]
    vim.command('vsplit')
    assert vim.windows[1].width == vim.windows[0].width // 2
    vim.windows[1].width = 2
    assert vim.windows[1].width == 2


def test_vars(vim: Nvim) -> None:
    vim.current.window.vars['python'] = [1, 2, {'3': 1}]
    assert vim.current.window.vars['python'] == [1, 2, {'3': 1}]
    assert vim.eval('w:python') == [1, 2, {'3': 1}]
    assert vim.current.window.vars.get('python') == [1, 2, {'3': 1}]

    del vim.current.window.vars['python']
    with pytest.raises(KeyError):
        vim.current.window.vars['python']
    assert vim.eval('exists("w:python")') == 0

    with pytest.raises(KeyError):
        del vim.current.window.vars['python']

    assert vim.current.window.vars.get('python', 'default') == 'default'


def test_options(vim: Nvim) -> None:
    vim.current.window.options['colorcolumn'] = '4,3'
    assert vim.current.window.options['colorcolumn'] == '4,3'
    old_global_statusline = vim.options['statusline']
    # global-local option
    vim.current.window.options['statusline'] = 'window-status'
    assert vim.current.window.options['statusline'] == 'window-status'
    assert vim.options['statusline'] == old_global_statusline

    with pytest.raises(KeyError) as excinfo:
        vim.current.window.options['doesnotexist']
    assert excinfo.value.args == ("Invalid option name: 'doesnotexist'",)


def test_position(vim: Nvim) -> None:
    height = vim.windows[0].height
    width = vim.windows[0].width
    vim.command('split')
    vim.command('vsplit')
    assert (vim.windows[0].row, vim.windows[0].col) == (0, 0)
    vsplit_pos = width / 2
    split_pos = height / 2
    assert vim.windows[1].row == 0
    assert vsplit_pos - 1 <= vim.windows[1].col <= vsplit_pos + 1
    assert split_pos - 1 <= vim.windows[2].row <= split_pos + 1
    assert vim.windows[2].col == 0


def test_tabpage(vim: Nvim) -> None:
    vim.command('tabnew')
    vim.command('vsplit')
    assert vim.windows[0].tabpage == vim.tabpages[0]
    assert vim.windows[1].tabpage == vim.tabpages[1]
    assert vim.windows[2].tabpage == vim.tabpages[1]


def test_valid(vim: Nvim) -> None:
    vim.command('split')
    window = vim.windows[1]
    vim.current.window = window
    assert window.valid
    vim.command('q')
    assert not window.valid


def test_number(vim: Nvim) -> None:
    curnum = vim.current.window.number
    vim.command('bot split')
    assert vim.current.window.number == curnum + 1
    vim.command('bot split')
    assert vim.current.window.number == curnum + 2


def test_handle(vim: Nvim) -> None:
    hnd1 = vim.current.window.handle
    vim.command('bot split')
    hnd2 = vim.current.window.handle
    assert hnd2 != hnd1
    vim.command('bot split')
    hnd3 = vim.current.window.handle
    assert hnd1 != hnd2 != hnd3
    vim.command('wincmd w')
    assert vim.current.window.handle == hnd1


def test_repr(vim: Nvim) -> None:
    assert repr(vim.current.window) == "<Window(handle=1000)>"



"""Wrapper to expose the Python interpreter as `pynvim-python`.

`setup.py` declares an entry point for the `main()` function below. When
`pynvim` is installed, an executable named `pynvim-python` will be generated
that will invoke `main()` below; that function then simply chains to the
underlying Python interpreter, passing along all command-line arguments.

The intent is to have `pynvim-python` be on the `PATH` such that an invocation
such as:

    pynvim-python -c 'import pynvim'

is equivalent to explicitly running the correct Python interpreter where
`pynvim` is installed:

    /path/to/python -c 'import pynvim'

This allows Neovim to automatically detect the correct Python interpreter for
use with `pynvim`.
"""

import subprocess
import sys


def main() -> None:
    """Chain to Python interpreter, passing all command-line args."""
    subprocess.run([sys.executable] + sys.argv[1:])



"""Event loop implementation that uses the `asyncio` standard module."""

from __future__ import annotations

import asyncio
import logging
import os
import sys
from collections import deque
from signal import Signals
from typing import Any, Callable, Deque, List, Optional, cast

if sys.version_info >= (3, 12):
    from typing import Final, override
else:
    from typing_extensions import Final, override

from pynvim.msgpack_rpc.event_loop.base import BaseEventLoop, TTransportType

logger = logging.getLogger(__name__)
debug, info, warn = (logger.debug, logger.info, logger.warning,)

loop_cls = asyncio.SelectorEventLoop

if os.name == 'nt':
    import msvcrt  # pylint: disable=import-error
    from asyncio.windows_utils import PipeHandle  # type: ignore[attr-defined]

    # On windows use ProactorEventLoop which support pipes and is backed by the
    # more powerful IOCP facility
    # NOTE: we override in the stdio case, because it doesn't work.
    loop_cls = asyncio.ProactorEventLoop  # type: ignore[attr-defined,misc]


# pylint: disable=logging-fstring-interpolation

class Protocol(asyncio.Protocol, asyncio.SubprocessProtocol):
    """The protocol class used for asyncio-based RPC communication."""

    def __init__(self, on_data, on_error):
        """Initialize the Protocol object."""
        assert on_data is not None
        assert on_error is not None
        self._on_data = on_data
        self._on_error = on_error

    @override
    def connection_made(self, transport):
        """Used to signal `asyncio.Protocol` of a successful connection."""
        self._transport = transport

    @override
    def connection_lost(self, exc: Optional[Exception]) -> None:
        """Used to signal `asyncio.Protocol` of a lost connection."""
        pass # replaces next logging statement
        # warn(f"connection_lost: exc = {exc}")

        self._on_error(exc if exc else EOFError("connection_lost"))

    @override
    def data_received(self, data: bytes) -> None:
        """Used to signal `asyncio.Protocol` of incoming data."""
        self._on_data(data)

    @override
    def pipe_connection_lost(self, fd: int, exc: Optional[Exception]) -> None:
        """Used to signal `asyncio.SubprocessProtocol` of a lost connection."""

        assert isinstance(self._transport, asyncio.SubprocessTransport)
        debug_info = {'fd': fd, 'exc': exc, 'pid': self._transport.get_pid()}
        pass # replaces next logging statement
        # warn(f"pipe_connection_lost {debug_info}")

        if os.name == 'nt' and fd == 2:  # stderr
            # On windows, ignore piped stderr being closed immediately (#505)
            return

        # pipe_connection_lost() *may* be called before process_exited() is
        # called, when a Nvim subprocess crashes (SIGABRT). Do not handle
        # errors here, as errors will be handled somewhere else
        # self._on_error(exc if exc else EOFError("pipe_connection_lost"))

    @override
    def pipe_data_received(self, fd, data):
        """Used to signal `asyncio.SubprocessProtocol` of incoming data."""
        if fd == 2:  # stderr fd number
            # Ignore stderr message, log only for debugging
            pass # replaces next logging statement
            # debug("stderr: %s", str(data))
        elif fd == 1:  # stdout
            self.data_received(data)

    @override
    def process_exited(self) -> None:
        """Used to signal `asyncio.SubprocessProtocol` when the child exits."""
        assert isinstance(self._transport, asyncio.SubprocessTransport)
        pid = self._transport.get_pid()
        return_code = self._transport.get_returncode()

        pass # replaces next logging statement
        # warn("process_exited, pid = %s, return_code = %s", pid, return_code)
        err = EOFError(f"process_exited: pid = {pid}, return_code = {return_code}")
        self._on_error(err)


class AsyncioEventLoop(BaseEventLoop):
    """`BaseEventLoop` subclass that uses core `asyncio` as a backend."""

    _protocol: Optional[Protocol]
    _transport: Optional[asyncio.WriteTransport]
    _signals: List[Signals]
    _data_buffer: Deque[bytes]
    if os.name != 'nt':
        _child_watcher: Optional[asyncio.AbstractChildWatcher]

    def __init__(self,
                 transport_type: TTransportType,
                 *args: Any, **kwargs: Any):
        """asyncio-specific initialization. see BaseEventLoop.__init__."""

        # The underlying asyncio event loop.
        self._loop: Final[asyncio.AbstractEventLoop] = loop_cls()

        # Handle messages from nvim that may arrive before run() starts.
        self._data_buffer = deque()

        def _on_data(data: bytes) -> None:
            if self._on_data is None:
                self._data_buffer.append(data)
                return
            self._on_data(data)

        # pylint: disable-next=unnecessary-lambda
        self._protocol_factory = lambda: Protocol(
            on_data=_on_data,
            on_error=self._on_error,
        )
        self._protocol = None

        # The communication channel (endpoint) created by _connect_*() methods,
        # where we write request messages to be sent to neovim
        self._transport = None
        self._to_close: List[asyncio.BaseTransport] = []
        self._child_watcher = None

        super().__init__(transport_type, *args, **kwargs)

    @override
    def _connect_tcp(self, address: str, port: int) -> None:
        async def connect_tcp():
            transport, protocol = await self._loop.create_connection(
                self._protocol_factory, address, port)
            pass # replaces next logging statement
            # debug(f"tcp connection successful: {address}:{port}")
            self._transport = transport
            self._protocol = protocol

        self._loop.run_until_complete(connect_tcp())

    @override
    def _connect_socket(self, path: str) -> None:
        async def connect_socket():
            if os.name == 'nt':
                _create_connection = self._loop.create_pipe_connection
            else:
                _create_connection = self._loop.create_unix_connection

            transport, protocol = await _create_connection(
                self._protocol_factory, path)
            pass # replaces next logging statement
            # debug("socket connection successful: %s", self._transport)
            self._transport = transport
            self._protocol = protocol

        self._loop.run_until_complete(connect_socket())

    @override
    def _connect_stdio(self) -> None:
        async def connect_stdin():
            if os.name == 'nt':
                pipe = PipeHandle(msvcrt.get_osfhandle(sys.stdin.fileno()))
            else:
                pipe = sys.stdin
            transport, protocol = await self._loop.connect_read_pipe(
                self._protocol_factory, pipe)
            pass # replaces next logging statement
            # debug("native stdin connection successful")
            self._to_close.append(transport)
            del protocol
        self._loop.run_until_complete(connect_stdin())

        # Make sure subprocesses don't clobber stdout,
        # send the output to stderr instead.
        rename_stdout = os.dup(sys.stdout.fileno())
        os.dup2(sys.stderr.fileno(), sys.stdout.fileno())

        async def connect_stdout():
            if os.name == 'nt':
                pipe = PipeHandle(msvcrt.get_osfhandle(rename_stdout))
            else:
                pipe = os.fdopen(rename_stdout, 'wb')

            transport, protocol = await self._loop.connect_write_pipe(
                self._protocol_factory, pipe)
            pass # replaces next logging statement
            # debug("native stdout connection successful")
            self._transport = transport
            self._protocol = protocol
        self._loop.run_until_complete(connect_stdout())

    @override
    def _connect_child(self, argv: List[str]) -> None:
        def get_child_watcher():
            try:
                return asyncio.get_child_watcher()
            except AttributeError:  # Python 3.14
                return None

            return None

        if os.name != 'nt':
            # see #238, #241
            watcher = get_child_watcher()
            if watcher is not None:
                watcher.attach_loop(self._loop)
                self._child_watcher = watcher

        async def create_subprocess():
            transport: asyncio.SubprocessTransport  # type: ignore
            transport, protocol = await self._loop.subprocess_exec(
                self._protocol_factory, *argv)
            pid = transport.get_pid()
            pass # replaces next logging statement
            # debug("child subprocess_exec successful, PID = %s", pid)

            self._transport = cast(asyncio.WriteTransport,
                                   transport.get_pipe_transport(0))  # stdin
            self._protocol = protocol

            # proactor transport implementations do not close the pipes
            # automatically, so make sure they are closed upon shutdown
            def _close_later(transport):
                if transport is not None:
                    self._to_close.append(transport)

            _close_later(transport.get_pipe_transport(1))
            _close_later(transport.get_pipe_transport(2))
            _close_later(transport)

        # await until child process have been launched and the transport has
        # been established
        self._loop.run_until_complete(create_subprocess())

    @override
    def _start_reading(self) -> None:
        pass

    @override
    def _send(self, data: bytes) -> None:
        assert self._transport, "connection has not been established."
        self._transport.write(data)

    @override
    def _run(self) -> None:
        # process the early messages that arrived as soon as the transport
        # channels are open and on_data is fully ready to receive messages.
        while self._data_buffer:
            data: bytes = self._data_buffer.popleft()
            if self._on_data is not None:
                self._on_data(data)

        self._loop.run_forever()

    @override
    def _stop(self) -> None:
        self._loop.stop()

    @override
    def _close(self) -> None:
        def _close_transport(transport):
            transport.close()

            # Windows: for ProactorBasePipeTransport, close() doesn't take in
            # effect immediately (closing happens asynchronously inside the
            # event loop), need to wait a bit for completing graceful shutdown.
            if (sys.version_info < (3, 13) and
                    os.name == 'nt' and hasattr(transport, '_sock')):
                async def wait_until_closed():
                    # pylint: disable-next=protected-access
                    while transport._sock is not None:
                        await asyncio.sleep(0.01)
                self._loop.run_until_complete(wait_until_closed())

        if self._transport:
            _close_transport(self._transport)
            self._transport = None
        for transport in self._to_close:
            _close_transport(transport)
        self._to_close[:] = []

        self._loop.close()

        if self._child_watcher is not None:
            self._child_watcher.close()
            self._child_watcher = None

    @override
    def _threadsafe_call(self, fn: Callable[[], Any]) -> None:
        self._loop.call_soon_threadsafe(fn)

    @override
    def _setup_signals(self, signals: List[Signals]) -> None:
        if os.name == 'nt':
            # add_signal_handler is not supported in win32
            self._signals = []
            return

        self._signals = list(signals)
        for signum in self._signals:
            self._loop.add_signal_handler(signum, self._on_signal, signum)

    @override
    def _teardown_signals(self) -> None:
        for signum in self._signals:
            self._loop.remove_signal_handler(signum)



"""Common code for event loop implementations."""
import logging
import signal
import sys
import threading
from abc import ABC, abstractmethod
from typing import Any, Callable, List, Optional, Union

if sys.version_info < (3, 8):
    from typing_extensions import Literal
else:
    from typing import Literal

logger = logging.getLogger(__name__)
debug, info, warn = (logger.debug, logger.info, logger.warning,)


# When signals are restored, the event loop library may reset SIGINT to SIG_DFL
# which exits the program. To be able to restore the python interpreter to it's
# default state, we keep a reference to the default handler
default_int_handler = signal.getsignal(signal.SIGINT)
main_thread = threading.current_thread()

TTransportType = Union[
    Literal['stdio'],
    Literal['socket'],
    Literal['tcp'],
    Literal['child']
]

# TODO: Since pynvim now supports python 3, the only available backend of the
# msgpack_rpc BaseEventLoop is the built-in asyncio (see #294). We will have
# to remove some unnecessary abstractions as well as greenlet. See also #489


class BaseEventLoop(ABC):
    """Abstract base class for all event loops.

    Event loops act as the bottom layer for Nvim sessions created by this
    library. They hide system/transport details behind a simple interface for
    reading/writing bytes to the connected Nvim instance.

    A lifecycle of event loop is as follows: (1. -> [2. -> 3.]* -> 4.)
      1. initialization (__init__): connection to Nvim is established.
      2. run(data_cb): run the event loop (blocks until the loop stops).
         Requests are sent to the remote neovim by calling send(), and
         responses (messages) from the remote neovim will be passed to the
         given `data_cb` callback function while the event loop is running.
         Note that run() may be called multiple times.
      3. stop(): stop the event loop.
      4. close(): close the event loop, destroying all the internal resources.

    This class exposes public methods for interacting with the underlying
    event loop and delegates implementation-specific work to the following
    methods, which subclasses are expected to implement:

    - `_init()`: Implementation-specific initialization
    - `_connect_tcp(address, port)`: connect to Nvim using tcp/ip
    - `_connect_socket(path)`: Same as tcp, but use a UNIX domain socket or
      named pipe.
    - `_connect_stdio()`: Use stdin/stdout as the connection to Nvim
    - `_connect_child(argv)`: Use the argument vector `argv` to spawn an
      embedded Nvim that has its stdin/stdout connected to the event loop.
    - `_start_reading()`: Called after any of _connect_* methods. Can be used
      to perform any post-connection setup or validation.
    - `_send(data)`: Send `data` (byte array) to Nvim (usually RPC request).
    - `_run()`: Runs the event loop until stopped or the connection is closed.
      The following methods can be called upon some events by the event loop:
      - `_on_data(data)`: When Nvim sends some data (usually RPC response).
      - `_on_signal(signum)`: When a signal is received.
      - `_on_error(exc)`: When a non-recoverable error occurs (e.g:
        connection lost, or any other OSError)
      Note that these _on_{data,signal,error} methods are not 'final', may be
      changed around an execution of run(). The subclasses are expected to
      handle any early messages arriving while _on_data is not yet set.
    - `_stop()`: Stop the event loop.
    - `_interrupt(data)`: Like `stop()`, but may be called from other threads
      this.
    - `_setup_signals(signals)`: Add implementation-specific listeners for
      for `signals`, which is a list of OS-specific signal numbers.
    - `_teardown_signals()`: Removes signal listeners set by `_setup_signals`
    """

    def __init__(self, transport_type: TTransportType, *args: Any, **kwargs: Any):
        """Initialize and connect the event loop instance.

        The only arguments are the transport type and transport-specific
        configuration, like this:

        >>> BaseEventLoop('tcp', '127.0.0.1', 7450)
        >>> BaseEventLoop('socket', '/tmp/nvim-socket')
        >>> BaseEventLoop('stdio')
        >>> BaseEventLoop('child', ['nvim', '--embed', '--headless', '-u', 'NONE'])

        Implementation-specific initialization should be made in the __init__
        constructor of the subclass, which must call the constructor of the
        super class (BaseEventLoop), in which one of the `_connect_*` methods
        (based on `transport_type`) and then `_start_reading()`.
        """
        self._transport_type = transport_type
        self._signames = dict((k, v) for v, k in signal.__dict__.items()
                              if v.startswith('SIG'))
        self._on_data: Optional[Callable[[bytes], None]] = None
        self._error: Optional[BaseException] = None
        try:
            getattr(self, '_connect_{}'.format(transport_type))(*args, **kwargs)
        except Exception as e:
            self.close()
            raise e
        self._start_reading()

    @abstractmethod
    def _start_reading(self) -> None:
        raise NotImplementedError()

    @abstractmethod
    def _send(self, data: bytes) -> None:
        raise NotImplementedError()

    def connect_tcp(self, address: str, port: int) -> None:
        """Connect to tcp/ip `address`:`port`. Delegated to `_connect_tcp`."""
        pass # replaces next logging statement
        # info('Connecting to TCP address: %s:%d', address, port)
        self._connect_tcp(address, port)

    @abstractmethod
    def _connect_tcp(self, address: str, port: int) -> None:
        raise NotImplementedError()

    def connect_socket(self, path: str) -> None:
        """Connect to socket at `path`. Delegated to `_connect_socket`."""
        pass # replaces next logging statement
        # info('Connecting to %s', path)
        self._connect_socket(path)

    @abstractmethod
    def _connect_socket(self, path: str) -> None:
        raise NotImplementedError()

    def connect_stdio(self) -> None:
        """Connect using stdin/stdout. Delegated to `_connect_stdio`."""
        pass # replaces next logging statement
        # info('Preparing stdin/stdout for streaming data')
        self._connect_stdio()

    @abstractmethod
    def _connect_stdio(self) -> None:
        raise NotImplementedError()

    def connect_child(self, argv):
        """Connect a new Nvim instance. Delegated to `_connect_child`."""
        pass # replaces next logging statement
        # info('Spawning a new nvim instance')
        self._connect_child(argv)

    @abstractmethod
    def _connect_child(self, argv: List[str]) -> None:
        raise NotImplementedError()

    def send(self, data: bytes) -> None:
        """Queue `data` for sending to Nvim."""
        pass # replaces next logging statement
        # debug("Sending '%s'", data)
        self._send(data)

    def threadsafe_call(self, fn):
        """Call a function in the event loop thread.

        This is the only safe way to interact with a session from other
        threads.
        """
        self._threadsafe_call(fn)

    @abstractmethod
    def _threadsafe_call(self, fn: Callable[[], Any]) -> None:
        raise NotImplementedError()

    def run(self, data_cb: Callable[[bytes], None]) -> None:
        """Run the event loop, and receives response messages to a callback."""
        if self._error:
            err = self._error
            if isinstance(self._error, KeyboardInterrupt):
                # KeyboardInterrupt is not destructive (it may be used in
                # the REPL).
                # After throwing KeyboardInterrupt, cleanup the _error field
                # so the loop may be started again
                self._error = None
            raise err

        # data_cb: e.g., MsgpackStream._on_data
        self._on_data = data_cb
        if threading.current_thread() == main_thread:
            self._setup_signals([signal.SIGINT, signal.SIGTERM])
        pass # replaces next logging statement
        # debug('Entering event loop')
        self._run()
        pass # replaces next logging statement
        # debug('Exited event loop')
        if threading.current_thread() == main_thread:
            self._teardown_signals()
            signal.signal(signal.SIGINT, default_int_handler)
        self._on_data = None

        # eventloop was stopped due to an error, re-raise it
        # (e.g. connection lost when subprocess nvim dies)
        if self._error:
            # Note: traceback is not preserved and attached for some reason,
            # should be somewhere from msgpack_rpc.event_loop.asyncio.Protocol
            raise self._error

    @abstractmethod
    def _run(self) -> None:
        raise NotImplementedError()

    def stop(self) -> None:
        """Stop the event loop."""
        self._stop()
        pass # replaces next logging statement
        # debug('Stopped event loop')

    @abstractmethod
    def _stop(self) -> None:
        raise NotImplementedError()

    def close(self) -> None:
        """Stop the event loop."""
        self._close()
        pass # replaces next logging statement
        # debug('Closed event loop')

    @abstractmethod
    def _close(self) -> None:
        raise NotImplementedError()

    def _on_signal(self, signum: signal.Signals) -> None:
        # pylint: disable-next=consider-using-f-string
        msg = 'Received signal {}'.format(self._signames[signum])
        pass # replaces next logging statement
        # debug(msg)

        if signum == signal.SIGINT and self._transport_type == 'stdio':
            # When the transport is stdio, we are probably running as a Nvim
            # child process. In that case, we don't want to be killed by
            # ctrl+C
            return

        if signum == signal.SIGINT:
            self._error = KeyboardInterrupt()
        else:
            self._error = Exception(msg)
        self.stop()

    def _on_error(self, exc: Exception) -> None:
        pass # replaces next logging statement
        # warn('on_error: %s', repr(exc))
        if self._error is None:
            # ignore subsequent exceptions, it's enough to raise only
            # the first exception arrived
            self._error = exc
        self.stop()

    def _on_interrupt(self) -> None:
        self.stop()

    def _setup_signals(self, signals: List[signal.Signals]) -> None:
        pass  # no-op by default

    def _teardown_signals(self) -> None:
        pass  # no-op by default



import os
import sys
from typing import Any


def test_setup_logging(monkeypatch: Any, tmpdir: str, caplog: Any) -> None:
    from pynvim import setup_logging

    major_version = sys.version_info[0]

    setup_logging('name1')
    assert caplog.messages == []

    def get_expected_logfile(prefix: str, name: str) -> str:
        return '{}_py{}_{}'.format(prefix, major_version, name)

    prefix = tmpdir.join('testlog1')
    monkeypatch.setenv('NVIM_PYTHON_LOG_FILE', str(prefix))
    setup_logging('name2')
    assert caplog.messages == []
    logfile = get_expected_logfile(prefix, 'name2')
    assert os.path.exists(logfile)
    assert open(logfile, 'r').read() == ''

    monkeypatch.setenv('NVIM_PYTHON_LOG_LEVEL', 'invalid')
    setup_logging('name3')
    assert caplog.record_tuples == [
        ('pynvim', 30, "Invalid NVIM_PYTHON_LOG_LEVEL: 'invalid', using INFO."),
    ]
    logfile = get_expected_logfile(prefix, 'name2')
    assert os.path.exists(logfile)
    with open(logfile, 'r') as f:
        lines = f.readlines()
        assert len(lines) == 1
        assert lines[0].endswith(
            "- Invalid NVIM_PYTHON_LOG_LEVEL: 'invalid', using INFO.\n"
        )



"""API for working with Nvim windows."""

from __future__ import annotations

from typing import TYPE_CHECKING, Tuple, cast

from pynvim.api.buffer import Buffer
from pynvim.api.common import Remote

if TYPE_CHECKING:
    from pynvim.api.tabpage import Tabpage


__all__ = ['Window']


class Window(Remote):

    """A remote Nvim window."""

    _api_prefix = "nvim_win_"

    @property
    def buffer(self) -> Buffer:
        """Get the `Buffer` currently being displayed by the window."""
        return self.request('nvim_win_get_buf')

    @property
    def cursor(self) -> Tuple[int, int]:
        """Get the (row, col) tuple with the current cursor position."""
        return cast(Tuple[int, int], tuple(self.request('nvim_win_get_cursor')))

    @cursor.setter
    def cursor(self, pos: Tuple[int, int]) -> None:
        """Set the (row, col) tuple as the new cursor position."""
        return self.request('nvim_win_set_cursor', pos)

    @property
    def height(self) -> int:
        """Get the window height in rows."""
        return self.request('nvim_win_get_height')

    @height.setter
    def height(self, height: int) -> None:
        """Set the window height in rows."""
        return self.request('nvim_win_set_height', height)

    @property
    def width(self) -> int:
        """Get the window width in rows."""
        return self.request('nvim_win_get_width')

    @width.setter
    def width(self, width: int) -> None:
        """Set the window height in rows."""
        return self.request('nvim_win_set_width', width)

    @property
    def row(self) -> int:
        """0-indexed, on-screen window position(row) in display cells."""
        return self.request('nvim_win_get_position')[0]

    @property
    def col(self) -> int:
        """0-indexed, on-screen window position(col) in display cells."""
        return self.request('nvim_win_get_position')[1]

    @property
    def tabpage(self) -> Tabpage:
        """Get the `Tabpage` that contains the window."""
        return self.request('nvim_win_get_tabpage')

    @property
    def valid(self) -> bool:
        """Return True if the window still exists."""
        return self.request('nvim_win_is_valid')

    @property
    def number(self) -> int:
        """Get the window number."""
        return self.request('nvim_win_get_number')



# type: ignore
from pynvim.plugin.decorators import command


def test_command_count() -> None:
    def function() -> None:
        """A dummy function to decorate."""
        return

    # ensure absence with default value of None
    decorated = command('test')(function)
    assert 'count' not in decorated._nvim_rpc_spec['opts']

    # ensure absence with explicit value of None
    count_value = None
    decorated = command('test', count=count_value)(function)
    assert 'count' not in decorated._nvim_rpc_spec['opts']

    # Test precedence with value of 0
    count_value = 0
    decorated = command('test', count=count_value)(function)
    assert 'count' in decorated._nvim_rpc_spec['opts']
    assert decorated._nvim_rpc_spec['opts']['count'] == count_value

    # Test presence with value of 1
    count_value = 1
    decorated = command('test', count=count_value)(function)
    assert 'count' in decorated._nvim_rpc_spec['opts']
    assert decorated._nvim_rpc_spec['opts']['count'] == count_value



from threading import Timer
from typing import List

from pynvim.api import Nvim


def test_interrupt_from_another_thread(vim: Nvim) -> None:
    timer = Timer(0.5, lambda: vim.async_call(lambda: vim.stop_loop()))
    timer.start()
    assert vim.next_message() is None


def test_exception_in_threadsafe_call(vim: Nvim) -> None:
    # an exception in a threadsafe_call shouldn't crash the entire host
    msgs: List[str] = []
    vim.async_call(
        lambda: [
            vim.eval("3"),
            undefined_variable  # type: ignore[name-defined] # noqa: F821
        ]
    )
    timer = Timer(0.5, lambda: vim.async_call(lambda: vim.stop_loop()))
    timer.start()
    vim.run_loop(None, None, err_cb=msgs.append)
    assert len(msgs) == 1
    msgs[0].index('NameError')
    msgs[0].index('undefined_variable')



import neovim

import pynvim.api


@neovim.plugin
class SimplePlugin:
    def __init__(self, nvim: pynvim.api.Nvim):
        self.nvim = nvim

    @neovim.command("SimpleHelloWorld")
    def hello_world(self) -> None:
        self.nvim.command("echom 'SimplePlugin: Hello World!'")



# -*- coding: utf-8 -*-
import time
from typing import List

from pynvim.api import Nvim


def test_call_and_reply(vim: Nvim) -> None:
    cid = vim.channel_id

    def setup_cb() -> None:
        cmd = 'let g:result = rpcrequest(%d, "client-call", 1, 2, 3)' % cid
        vim.command(cmd)
        assert vim.vars['result'] == [4, 5, 6]
        vim.stop_loop()

    def request_cb(name: str, args: List[int]) -> List[int]:
        assert name == 'client-call'
        assert args == [1, 2, 3]
        return [4, 5, 6]

    vim.run_loop(request_cb, None, setup_cb)


def test_call_api_before_reply(vim: Nvim) -> None:
    cid = vim.channel_id

    def setup_cb() -> None:
        cmd = 'let g:result = rpcrequest(%d, "client-call2", 1, 2, 3)' % cid
        vim.command(cmd)
        assert vim.vars['result'] == [7, 8, 9]
        vim.stop_loop()

    def request_cb(name: str, args: List[int]) -> List[int]:
        vim.command('let g:result2 = [7, 8, 9]')
        return vim.vars['result2']

    vim.run_loop(request_cb, None, setup_cb)


def test_async_call(vim: Nvim) -> None:

    def request_cb(name: str, args: List[int]) -> None:
        if name == "test-event":
            vim.vars['result'] = 17
        vim.stop_loop()

    # this would have dead-locked if not async
    vim.funcs.rpcrequest(vim.channel_id, "test-event", async_=True)
    vim.run_loop(request_cb, None, None)

    # TODO(blueyed): This sleep is required on Travis, where it hangs with
    # "Entering event loop" otherwise  (asyncio's EpollSelector._epoll.poll).
    time.sleep(0.1)

    assert vim.vars['result'] == 17


def test_recursion(vim: Nvim) -> None:
    cid = vim.channel_id

    def setup_cb() -> None:
        vim.vars['result1'] = 0
        vim.vars['result2'] = 0
        vim.vars['result3'] = 0
        vim.vars['result4'] = 0
        cmd = 'let g:result1 = rpcrequest(%d, "call", %d)' % (cid, 2,)
        vim.command(cmd)
        assert vim.vars['result1'] == 4
        assert vim.vars['result2'] == 8
        assert vim.vars['result3'] == 16
        assert vim.vars['result4'] == 32
        vim.stop_loop()

    def request_cb(name: str, args: List[int]) -> int:
        n = args[0]
        n *= 2
        if n <= 16:
            if n == 4:
                cmd = 'let g:result2 = rpcrequest(%d, "call", %d)' % (cid, n,)
            elif n == 8:
                cmd = 'let g:result3 = rpcrequest(%d, "call", %d)' % (cid, n,)
            elif n == 16:
                cmd = 'let g:result4 = rpcrequest(%d, "call", %d)' % (cid, n,)
            vim.command(cmd)
        return n

    vim.run_loop(request_cb, None, setup_cb)



import pytest

from pynvim.api import Nvim


def test_windows(vim: Nvim) -> None:
    vim.command('tabnew')
    vim.command('vsplit')
    assert list(vim.tabpages[0].windows) == [vim.windows[0]]
    assert list(vim.tabpages[1].windows) == [vim.windows[1], vim.windows[2]]
    assert vim.tabpages[1].window == vim.windows[1]
    vim.current.window = vim.windows[2]
    assert vim.tabpages[1].window == vim.windows[2]


def test_vars(vim: Nvim) -> None:
    vim.current.tabpage.vars['python'] = [1, 2, {'3': 1}]
    assert vim.current.tabpage.vars['python'] == [1, 2, {'3': 1}]
    assert vim.eval('t:python') == [1, 2, {'3': 1}]
    assert vim.current.tabpage.vars.get('python') == [1, 2, {'3': 1}]

    del vim.current.tabpage.vars['python']
    with pytest.raises(KeyError):
        vim.current.tabpage.vars['python']
    assert vim.eval('exists("t:python")') == 0

    with pytest.raises(KeyError):
        del vim.current.tabpage.vars['python']

    assert vim.current.tabpage.vars.get('python', 'default') == 'default'


def test_valid(vim: Nvim) -> None:
    vim.command('tabnew')
    tabpage = vim.tabpages[1]
    assert tabpage.valid
    vim.command('tabclose')
    assert not tabpage.valid


def test_number(vim: Nvim) -> None:
    curnum = vim.current.tabpage.number
    vim.command('tabnew')
    assert vim.current.tabpage.number == curnum + 1
    vim.command('tabnew')
    assert vim.current.tabpage.number == curnum + 2


def test_repr(vim: Nvim) -> None:
    assert repr(vim.current.tabpage) == "<Tabpage(handle=1)>"



"""The `mymodule` package for the fixture module plugin."""
# pylint: disable=all

# Somehow the plugin might be using relative imports.
from .plugin import MyPlugin as MyPlugin

# ... or absolute import (assuming this is the root package)
import mymodule.plugin  # noqa: I100
assert mymodule.plugin.MyPlugin is MyPlugin



"""Configs for pytest."""

import gc
import json
import os
import sys
from typing import Generator

import pytest

import pynvim

pynvim.setup_logging("test")


@pytest.fixture
def vim() -> Generator[pynvim.Nvim, None, None]:
    """Create an embedded, sub-process Nvim fixture instance."""
    editor: pynvim.Nvim

    child_argv = os.environ.get('NVIM_CHILD_ARGV')
    listen_address = os.environ.get('NVIM')
    if child_argv is None and listen_address is None:
        child_argv = json.dumps([
            "nvim",
            "--clean",  # no config and plugins (-u NONE), no SHADA
            "-n",  # no swap file
            "--embed",
            "--headless",
            # Always use the same exact python executable regardless of $PATH
            "--cmd", f"let g:python3_host_prog='{sys.executable}'",
        ])

    if child_argv is not None:
        editor = pynvim.attach('child', argv=json.loads(child_argv))
    else:
        assert listen_address is not None and listen_address != ''
        editor = pynvim.attach('socket', path=listen_address)

    try:
        yield editor

    finally:
        # Ensure all internal resources (pipes, transports, etc.) are always
        # closed properly. Otherwise, during GC finalizers (__del__) will raise
        # "Event loop is closed" error.
        editor.close()

        gc.collect()  # force-run GC, to early-detect potential leakages



"""setup.py for pynvim."""

import os.path
import platform
import sys

__PATH__ = os.path.abspath(os.path.dirname(__file__))

from setuptools import setup

install_requires = [
    'msgpack>=1.0.0',
    'greenlet>=3.0; python_implementation != "PyPy"',
    'typing-extensions>=4.5; python_version < "3.12"',
]

needs_pytest = {'pytest', 'test', 'ptr'}.intersection(sys.argv)

setup_requires = [
]

tests_require = [
    'pytest',
    'pytest_timeout',
]

docs_require = [
    'sphinx',
    'sphinx-rtd-theme',
]

extras_require = {
    'test': tests_require,
    'docs': docs_require,
}


# __version__: see pynvim/_version.py
with open(os.path.join(__PATH__, "pynvim/_version.py"),
          "r", encoding="utf-8") as fp:
    _version_env = {}
    exec(fp.read(), _version_env)  # pylint: disable=exec-used
    version = _version_env['__version__']


setup(name='pynvim',
      version=version,
      description='Python client for Neovim',
      url='http://github.com/neovim/pynvim',
      download_url=f'https://github.com/neovim/pynvim/archive/{version}.tar.gz',
      author='Neovim Authors',
      license='Apache',
      packages=['pynvim', 'pynvim.api', 'pynvim.msgpack_rpc',
                'pynvim.msgpack_rpc.event_loop', 'pynvim.plugin',
                'neovim', 'neovim.api'],
      python_requires=">=3.7",
      install_requires=install_requires,
      setup_requires=setup_requires,
      tests_require=tests_require,
      extras_require=extras_require,
      entry_points={
              'console_scripts': [
                  'pynvim-python=pynvim.python:main',
              ],
          },
      )



# type: ignore
"""Implements a Nvim host for python plugins."""

import importlib
import inspect
import logging
import os
import os.path
import pathlib
import re
import sys
from functools import partial
from traceback import format_exc
from typing import Any, Sequence

from pynvim.api import Nvim, decode_if_bytes, walk
from pynvim.msgpack_rpc import ErrorResponse
from pynvim.plugin import script_host
from pynvim.util import format_exc_skip, get_client_info

__all__ = ('Host',)

logger = logging.getLogger(__name__)
error, debug, info, warn = (logger.error, logger.debug, logger.info,
                            logger.warning,)

host_method_spec = {"poll": {}, "specs": {"nargs": 1}, "shutdown": {}}


def _handle_import(path: str, name: str):
    """Import python module `name` from a known file path or module directory.

    The path should be the base directory from which the module can be imported.
    To support python 3.12, the use of `imp` should be avoided.
    @see https://docs.python.org/3.12/whatsnew/3.12.html#imp
    """
    if not name:
        raise ValueError("Missing module name.")

    sys.path.append(path)
    return importlib.import_module(name)


class Host(object):

    """Nvim host for python plugins.

    Takes care of loading/unloading plugins and routing msgpack-rpc
    requests/notifications to the appropriate handlers.
    """

    def __init__(self, nvim: Nvim):
        """Set handlers for plugin_load/plugin_unload."""
        self.nvim = nvim
        self._specs = {}
        self._loaded = {}
        self._load_errors = {}
        self._notification_handlers = {
            'nvim_error_event': self._on_error_event
        }
        self._request_handlers = {
            'poll': lambda: 'ok',
            'specs': self._on_specs_request,
            'shutdown': self.shutdown
        }

        self._decode_default = True

    def _on_async_err(self, msg: str) -> None:
        # uncaught python exception
        self.nvim.err_write(msg, async_=True)

    def _on_error_event(self, kind: Any, msg: str) -> None:
        # error from nvim due to async request
        # like nvim.command(..., async_=True)
        errmsg = "{}: Async request caused an error:\n{}\n".format(
            self.name, decode_if_bytes(msg))
        self.nvim.err_write(errmsg, async_=True)
        return errmsg

    def start(self, plugins):
        """Start listening for msgpack-rpc requests and notifications."""
        self.nvim.run_loop(self._on_request,
                           self._on_notification,
                           lambda: self._load(plugins),
                           err_cb=self._on_async_err)

    def shutdown(self) -> None:
        """Shutdown the host."""
        self._unload()
        self.nvim.stop_loop()

    def _wrap_delayed_function(self, cls, delayed_handlers, name, sync,
                               module_handlers, path, *args):
        # delete the delayed handlers to be sure
        for handler in delayed_handlers:
            method_name = handler._nvim_registered_name
            if handler._nvim_rpc_sync:
                del self._request_handlers[method_name]
            else:
                del self._notification_handlers[method_name]
        # create an instance of the plugin and pass the nvim object
        plugin = cls(self._configure_nvim_for(cls))

        # discover handlers in the plugin instance
        self._discover_functions(plugin, module_handlers, path, False)

        if sync:
            return self._request_handlers[name](*args)
        else:
            return self._notification_handlers[name](*args)

    def _wrap_function(self, fn, sync, decode, nvim_bind, name, *args):
        if decode:
            args = walk(partial(decode_if_bytes, mode=decode), args)
        if nvim_bind is not None:
            args.insert(0, nvim_bind)
        try:
            return fn(*args)
        except Exception:
            if sync:
                msg = ("error caught in request handler '{} {}':\n{}"
                       .format(name, args, format_exc_skip(1)))
                raise ErrorResponse(msg)
            else:
                msg = ("error caught in async handler '{} {}'\n{}\n"
                       .format(name, args, format_exc_skip(1)))
                self._on_async_err(msg + "\n")

    def _on_request(self, name: str, args: Sequence[Any]) -> None:
        """Handle a msgpack-rpc request."""
        name = decode_if_bytes(name)
        handler = self._request_handlers.get(name, None)
        if not handler:
            msg = self._missing_handler_error(name, 'request')
            pass # replaces next logging statement
            # error(msg)
            raise ErrorResponse(msg)

        pass # replaces next logging statement
        # debug('calling request handler for "%s", args: "%s"', name, args)
        rv = handler(*args)
        pass # replaces next logging statement
        # debug("request handler for '%s %s' returns: %s", name, args, rv)
        return rv

    def _on_notification(self, name: str, args: Sequence[Any]) -> None:
        """Handle a msgpack-rpc notification."""
        name = decode_if_bytes(name)
        handler = self._notification_handlers.get(name, None)
        if not handler:
            msg = self._missing_handler_error(name, 'notification')
            pass # replaces next logging statement
            # error(msg)
            self._on_async_err(msg + "\n")
            return

        pass # replaces next logging statement
        # debug('calling notification handler for "%s", args: "%s"', name, args)
        handler(*args)

    def _missing_handler_error(self, name, kind):
        msg = 'no {} handler registered for "{}"'.format(kind, name)
        pathmatch = re.match(r'(.+):[^:]+:[^:]+', name)
        if pathmatch:
            loader_error = self._load_errors.get(pathmatch.group(1))
            if loader_error is not None:
                msg = msg + "\n" + loader_error
        return msg

    def _load(self, plugins: Sequence[str]) -> None:
        """Load the remote plugins and register handlers defined in the plugins.

        Args:
            plugins: List of plugin paths to rplugin python modules
                registered by remote#host#RegisterPlugin('python3', ...)
                (see the generated rplugin.vim manifest)
        """
        # self.nvim.err_write("host init _load\n", async_=True)
        has_script = False
        for path in plugins:
            path = pathlib.Path(os.path.normpath(path)).as_posix()  # normalize path
            err = None
            if path in self._loaded:
                pass # replaces next logging statement
                # warn('{} is already loaded'.format(path))
                continue
            try:
                if path == "script_host.py":
                    module = script_host
                    has_script = True
                else:
                    directory, name = os.path.split(os.path.splitext(path)[0])
                    module = _handle_import(directory, name)
                handlers = []
                self._discover_classes(module, handlers, path)
                self._discover_functions(module, handlers, path, False)
                if not handlers:
                    pass # replaces next logging statement
                    # error('{} exports no handlers'.format(path))
                    continue
                self._loaded[path] = {'handlers': handlers, 'module': module}
            except Exception as e:
                err = ('Encountered {} loading plugin at {}: {}\n{}'
                       .format(type(e).__name__, path, e, format_exc(5)))
                pass # replaces next logging statement
                # error(err)
                self._load_errors[path] = err

        kind = ("script-host" if len(plugins) == 1 and has_script
                else "rplugin-host")
        info = get_client_info(kind, 'host', host_method_spec)
        self.name = info[0]
        self.nvim.api.set_client_info(*info, async_=True)

    def _unload(self) -> None:
        for path, plugin in self._loaded.items():
            handlers = plugin['handlers']
            for handler in handlers:
                method_name = handler._nvim_registered_name
                if hasattr(handler, '_nvim_shutdown_hook'):
                    handler()
                elif handler._nvim_rpc_sync:
                    del self._request_handlers[method_name]
                else:
                    del self._notification_handlers[method_name]
        self._specs = {}
        self._loaded = {}

    def _discover_classes(self, module, handlers, plugin_path):
        for _, cls in inspect.getmembers(module, inspect.isclass):
            if getattr(cls, '_nvim_plugin', False):
                # discover handlers in the plugin instance
                self._discover_functions(cls, handlers, plugin_path, True)

    def _discover_functions(self, obj, handlers, plugin_path, delay):
        def predicate(o):
            return hasattr(o, '_nvim_rpc_method_name')

        cls_handlers = []
        specs = []
        objdecode = getattr(obj, '_nvim_decode', self._decode_default)
        for _, fn in inspect.getmembers(obj, predicate):
            method = fn._nvim_rpc_method_name
            if fn._nvim_prefix_plugin_path:
                method = '{}:{}'.format(plugin_path, method)
            sync = fn._nvim_rpc_sync
            if delay:
                fn_wrapped = partial(self._wrap_delayed_function, obj,
                                     cls_handlers, method, sync,
                                     handlers, plugin_path)
            else:
                decode = getattr(fn, '_nvim_decode', objdecode)
                nvim_bind = None
                if fn._nvim_bind:
                    nvim_bind = self._configure_nvim_for(fn)

                fn_wrapped = partial(self._wrap_function, fn,
                                     sync, decode, nvim_bind, method)
            self._copy_attributes(fn, fn_wrapped)
            fn_wrapped._nvim_registered_name = method
            # register in the rpc handler dict
            if sync:
                if method in self._request_handlers:
                    raise Exception(('Request handler for "{}" is '
                                     + 'already registered').format(method))
                self._request_handlers[method] = fn_wrapped
            else:
                if method in self._notification_handlers:
                    raise Exception(('Notification handler for "{}" is '
                                     + 'already registered').format(method))
                self._notification_handlers[method] = fn_wrapped
            if hasattr(fn, '_nvim_rpc_spec'):
                specs.append(fn._nvim_rpc_spec)
            handlers.append(fn_wrapped)
            cls_handlers.append(fn_wrapped)
        if specs:
            self._specs[plugin_path] = specs

    def _copy_attributes(self, fn, fn2):
        # Copy _nvim_* attributes from the original function
        for attr in dir(fn):
            if attr.startswith('_nvim_'):
                setattr(fn2, attr, getattr(fn, attr))

    def _on_specs_request(self, path):
        path = decode_if_bytes(path)
        path = pathlib.Path(os.path.normpath(path)).as_posix()  # normalize path
        if path in self._load_errors:
            self.nvim.out_write(self._load_errors[path] + '\n')
        return self._specs.get(path, 0)

    def _configure_nvim_for(self, obj):
        # Configure a nvim instance for obj (checks encoding configuration)
        nvim = self.nvim
        decode = getattr(obj, '_nvim_decode', self._decode_default)
        if decode:
            nvim = nvim.with_decode(decode)
        return nvim



"""Tests interaction with neovim via Nvim API (with child process)."""

import os
import sys
import tempfile
from pathlib import Path

import pytest

from pynvim.api import Nvim, NvimError


def source(vim: Nvim, code: str) -> None:
    fd, fname = tempfile.mkstemp()
    with os.fdopen(fd, 'w') as f:
        f.write(code)
    vim.command('source ' + fname)
    os.unlink(fname)


def test_clientinfo(vim: Nvim) -> None:
    assert 'remote' == vim.api.get_chan_info(vim.channel_id)['client']['type']


def test_command(vim: Nvim) -> None:
    fname = tempfile.mkstemp()[1]
    vim.command('new')
    vim.command('edit {}'.format(fname))
    # skip the "press return" state, which does not handle deferred calls
    vim.input('\r')
    vim.command('normal itesting\npython\napi')
    vim.command('w')
    assert os.path.isfile(fname)
    with open(fname) as f:
        assert f.read() == 'testing\npython\napi\n'
    try:
        os.unlink(fname)
    except OSError:
        pass  # on windows, this can be flaky; ignore it


def test_command_output(vim: Nvim) -> None:
    assert vim.command_output('echo "test"') == 'test'

    # can capture multi-line outputs
    vim.command("let g:multiline_string = join(['foo', 'bar'], nr2char(10))")
    assert vim.command_output('echo g:multiline_string') == "foo\nbar"


def test_command_error(vim: Nvim) -> None:
    with pytest.raises(vim.error) as excinfo:
        vim.current.window.cursor = -1, -1
    assert excinfo.value.args == ('Cursor position outside buffer',)


def test_eval(vim: Nvim) -> None:
    vim.command('let g:v1 = "a"')
    vim.command('let g:v2 = [1, 2, {"v3": 3}]')
    g = vim.eval('g:')
    assert g['v1'] == 'a'
    assert g['v2'] == [1, 2, {'v3': 3}]


def test_call(vim: Nvim) -> None:
    assert vim.funcs.join(['first', 'last'], ', ') == 'first, last'
    source(vim, """
        function! Testfun(a,b)
            return string(a:a).":".a:b
        endfunction
    """)
    assert vim.funcs.Testfun(3, 'alpha') == '3:alpha'


def test_api(vim: Nvim) -> None:
    vim.api.command('let g:var = 3')
    assert vim.api.eval('g:var') == 3


def test_strwidth(vim: Nvim) -> None:
    assert vim.strwidth('abc') == 3
    # 6 + (neovim)
    # 19 * 2 (each japanese character occupies two cells)
    assert vim.strwidth('neovimのデザインかなりまともなのになってる。') == 44


def test_chdir(vim: Nvim) -> None:
    pwd = vim.eval('getcwd()')
    root = os.path.abspath(os.sep)
    # We can chdir to '/' on Windows, but then the pwd will be the root drive
    vim.chdir('/')
    assert vim.eval('getcwd()') == root
    vim.chdir(pwd)
    assert vim.eval('getcwd()') == pwd


def test_current_line(vim: Nvim) -> None:
    assert vim.current.line == ''
    vim.current.line = 'abc'
    assert vim.current.line == 'abc'


def test_current_line_delete(vim: Nvim) -> None:
    vim.current.buffer[:] = ['one', 'two']
    assert len(vim.current.buffer[:]) == 2
    del vim.current.line
    assert len(vim.current.buffer[:]) == 1 and vim.current.buffer[0] == 'two'
    del vim.current.line
    assert len(vim.current.buffer[:]) == 1 and not vim.current.buffer[0]


def test_vars(vim: Nvim) -> None:
    vim.vars['python'] = [1, 2, {'3': 1}]
    assert vim.vars['python'] == [1, 2, {'3': 1}]
    assert vim.eval('g:python') == [1, 2, {'3': 1}]
    assert vim.vars.get('python') == [1, 2, {'3': 1}]

    del vim.vars['python']
    with pytest.raises(KeyError):
        vim.vars['python']
    assert vim.eval('exists("g:python")') == 0

    with pytest.raises(KeyError):
        del vim.vars['python']

    assert vim.vars.get('python', 'default') == 'default'


def test_options(vim: Nvim) -> None:
    assert vim.options['background'] == 'dark'
    vim.options['background'] = 'light'
    assert vim.options['background'] == 'light'


def test_local_options(vim: Nvim) -> None:
    assert vim.windows[0].options['foldmethod'] == 'manual'
    vim.windows[0].options['foldmethod'] = 'syntax'
    assert vim.windows[0].options['foldmethod'] == 'syntax'


def test_buffers(vim: Nvim) -> None:
    buffers = []

    # Number of elements
    assert len(vim.buffers) == 1

    # Indexing (by buffer number)
    assert vim.buffers[vim.current.buffer.number] == vim.current.buffer

    buffers.append(vim.current.buffer)
    vim.command('new')
    assert len(vim.buffers) == 2
    buffers.append(vim.current.buffer)
    assert vim.buffers[vim.current.buffer.number] == vim.current.buffer
    vim.current.buffer = buffers[0]
    assert vim.buffers[vim.current.buffer.number] == buffers[0]

    # Membership test
    assert buffers[0] in vim.buffers
    assert buffers[1] in vim.buffers
    assert {} not in vim.buffers  # type: ignore[operator]

    # Iteration
    assert buffers == list(vim.buffers)


def test_windows(vim: Nvim) -> None:
    assert len(vim.windows) == 1
    assert vim.windows[0] == vim.current.window
    vim.command('vsplit')
    vim.command('split')
    assert len(vim.windows) == 3
    assert vim.windows[0] == vim.current.window
    vim.current.window = vim.windows[1]
    assert vim.windows[1] == vim.current.window


def test_tabpages(vim: Nvim) -> None:
    assert len(vim.tabpages) == 1
    assert vim.tabpages[0] == vim.current.tabpage
    vim.command('tabnew')
    assert len(vim.tabpages) == 2
    assert len(vim.windows) == 2
    assert vim.windows[1] == vim.current.window
    assert vim.tabpages[1] == vim.current.tabpage
    vim.current.window = vim.windows[0]
    # Switching window also switches tabpages if necessary(this probably
    # isn't the current behavior, but compatibility will be handled in the
    # python client with an optional parameter)
    assert vim.tabpages[0] == vim.current.tabpage
    assert vim.windows[0] == vim.current.window
    vim.current.tabpage = vim.tabpages[1]
    assert vim.tabpages[1] == vim.current.tabpage
    assert vim.windows[1] == vim.current.window


def test_hash(vim: Nvim) -> None:
    d = {}
    d[vim.current.buffer] = "alpha"
    assert d[vim.current.buffer] == 'alpha'
    vim.command('new')
    d[vim.current.buffer] = "beta"
    assert d[vim.current.buffer] == 'beta'
    vim.command('winc w')
    assert d[vim.current.buffer] == 'alpha'
    vim.command('winc w')
    assert d[vim.current.buffer] == 'beta'


def test_python3(vim: Nvim) -> None:
    """Tests whether python3 host can load."""
    rv = vim.exec_lua('''
        local prog, err = vim.provider.python.detect_by_module("neovim")
        return { prog = prog, err = err }''')
    assert rv['prog'] != "", rv['err']
    assert rv['prog'] == sys.executable

    assert sys.executable == vim.command_output(
        'python3 import sys; print(sys.executable)')

    assert 1 == vim.eval('has("python3")')


def test_python3_ex_eval(vim: Nvim) -> None:
    assert '42' == vim.command_output('python3 =42')
    assert '42' == vim.command_output('python3 =   42     ')
    assert '42' == vim.command_output('py3=    42     ')
    assert '42' == vim.command_output('py=42')

    # On syntax error or evaluation error, stacktrace information is printed
    # Note: the pynvim API command_output() throws an exception on error
    # because the Ex command :python will throw (wrapped with provider#python3#Call)
    with pytest.raises(NvimError) as excinfo:
        vim.command('py3= 1/0')
    stacktrace = excinfo.value.args[0]
    assert 'File "<string>", line 1, in <module>' in stacktrace
    assert 'ZeroDivisionError: division by zero' in stacktrace

    vim.command('python3 def raise_error(): raise RuntimeError("oops")')
    with pytest.raises(NvimError) as excinfo:
        vim.command_output('python3 =print("nooo", raise_error())')
    stacktrace = excinfo.value.args[0]
    assert 'File "<string>", line 1, in raise_error' in stacktrace
    assert 'RuntimeError: oops' in stacktrace
    assert 'nooo' not in vim.command_output(':messages')


def test_python_cwd(vim: Nvim, tmp_path: Path) -> None:
    vim.command('python3 import os')
    cwd_before = vim.command_output('python3 print(os.getcwd())')

    # handle DirChanged #296
    vim.command('cd {}'.format(str(tmp_path)))
    cwd_vim = vim.command_output('pwd')
    cwd_python = vim.command_output('python3 print(os.getcwd())')
    assert cwd_python == cwd_vim
    assert cwd_python != cwd_before


lua_code = """
local a = vim.api
local y = ...
function pynvimtest_func(x)
    return x+y
end

local function setbuf(buf,lines)
   a.nvim_buf_set_lines(buf, 0, -1, true, lines)
end


local function getbuf(buf)
   return a.nvim_buf_line_count(buf)
end

pynvimtest = {setbuf=setbuf, getbuf=getbuf}

return "eggspam"
"""


def test_lua(vim: Nvim) -> None:
    assert vim.exec_lua(lua_code, 7) == "eggspam"
    assert vim.lua.pynvimtest_func(3) == 10
    lua_module = vim.lua.pynvimtest
    buf = vim.current.buffer
    lua_module.setbuf(buf, ["a", "b", "c", "d"], async_=True)
    assert lua_module.getbuf(buf) == 4



import pytest

from pynvim.api import Nvim


def test_receiving_events(vim: Nvim) -> None:
    vim.command('call rpcnotify(%d, "test-event", 1, 2, 3)' % vim.channel_id)
    event = vim.next_message()
    assert event[1] == 'test-event'
    assert event[2] == [1, 2, 3]
    vim.command('au FileType python call rpcnotify(%d, "py!", bufnr("$"))' %
                vim.channel_id)
    vim.command('set filetype=python')
    event = vim.next_message()
    assert event[1] == 'py!'
    assert event[2] == [vim.current.buffer.number]


def test_sending_notify(vim: Nvim) -> None:
    # notify after notify
    vim.command("let g:test = 3", async_=True)
    cmd = 'call rpcnotify(%d, "test-event", g:test)' % vim.channel_id
    vim.command(cmd, async_=True)
    event = vim.next_message()
    assert event[1] == 'test-event'
    assert event[2] == [3]

    # request after notify
    vim.command("let g:data = 'xyz'", async_=True)
    assert vim.eval('g:data') == 'xyz'


def test_async_error(vim: Nvim) -> None:
    # Invoke a bogus Ex command via notify (async).
    vim.command("lolwut", async_=True)
    event = vim.next_message()
    assert event[1] == 'nvim_error_event'


def test_broadcast(vim: Nvim) -> None:
    if (vim.version.major, vim.version.minor) < (0, 11):
        # see #570, neovim/neovim#28487
        pytest.skip("neovim/neovim#28487")

    vim.command('call rpcnotify(0, "event1", 1, 2, 3)')
    vim.command('call rpcnotify(0, "event2", 4, 5, 6)')
    vim.command('call rpcnotify(0, "event2", 7, 8, 9)')
    event = vim.next_message()
    assert event[1] == 'event1'
    assert event[2] == [1, 2, 3]
    event = vim.next_message()
    assert event[1] == 'event2'
    assert event[2] == [4, 5, 6]
    vim.command('call rpcnotify(0, "event2", 10, 11, 12)')
    vim.command('call rpcnotify(0, "event1", 13, 14, 15)')
    msg = vim.next_message()
    assert msg[1] == 'event2'
    assert msg[2] == [7, 8, 9]



"""Tests other session_types than subprocess Nvim."""

import contextlib
import os.path
import socket
import subprocess
import tempfile
import time
from typing import Generator

import pytest
import pytest_timeout  # pylint: disable=unused-import  # noqa

import pynvim
from pynvim.api import Nvim

# pylint: disable=consider-using-with
# pylint: disable=redefined-outer-name


xfail_on_windows = pytest.mark.xfail(
    "os.name == 'nt'", reason="Broken in Windows, see #544")


@pytest.fixture
def tmp_socket() -> Generator[str, None, None]:
    """Get a temporary UNIX socket file."""
    # see cpython#93914
    addr = tempfile.mktemp(prefix="test_python_", suffix='.sock',
                           dir=os.path.curdir)
    try:
        yield addr
    finally:
        if os.path.exists(addr):
            with contextlib.suppress(OSError):
                os.unlink(addr)


@xfail_on_windows
def test_connect_socket(tmp_socket: str) -> None:
    """Tests UNIX socket connection."""
    p = subprocess.Popen(["nvim", "--clean", "-n", "--headless",
                          "--listen", tmp_socket])
    time.sleep(0.2)  # wait a bit until nvim starts up

    try:
        nvim: Nvim = pynvim.attach('socket', path=tmp_socket)
        assert 42 == nvim.eval('42')
        assert "?" == nvim.command_output('echo "?"')
    finally:
        with contextlib.suppress(OSError):
            p.terminate()


def test_connect_socket_fail() -> None:
    """Tests UNIX socket connection, when the sock file is not found."""
    with pytest.raises(FileNotFoundError):
        pynvim.attach('socket', path='/tmp/not-exist.socket')


def find_free_port() -> int:
    """Find a free, available port number."""
    with socket.socket() as sock:
        sock.bind(('', 0))  # Bind to a free port provided by the host.
        return sock.getsockname()[1]


def test_connect_tcp() -> None:
    """Tests TCP connection."""
    address = '127.0.0.1'
    port = find_free_port()
    p = subprocess.Popen(["nvim", "--clean", "-n", "--headless",
                          "--listen", f"{address}:{port}"])
    time.sleep(0.2)  # wait a bit until nvim starts up

    try:
        nvim: Nvim = pynvim.attach('tcp', address=address, port=port)
        assert 42 == nvim.eval('42')
        assert "?" == nvim.command_output('echo "?"')
    finally:
        with contextlib.suppress(OSError):
            p.terminate()


@pytest.mark.timeout(5.0)
def test_connect_tcp_no_server() -> None:
    """Tests TCP socket connection that fails; connection refused."""
    port = find_free_port()

    with pytest.raises(ConnectionRefusedError):
        pynvim.attach('tcp', address='127.0.0.1', port=port)


@xfail_on_windows
def test_connect_stdio(vim: Nvim) -> None:
    """Tests stdio connection, using jobstart(..., {'rpc': v:true})."""

    def source(vim: Nvim, code: str) -> None:
        """Source a vimscript code in the embedded nvim instance."""
        fd, fname = tempfile.mkstemp()
        try:
            with os.fdopen(fd, 'w') as f:
                f.write(code)
            vim.command('source ' + fname)
        finally:
            os.unlink(fname)

    # A helper function for debugging that captures what pynvim writes to
    # stderr (e.g. python stacktrace): used as a |on_stderr| callback
    source(vim, """
        function! OutputHandler(j, lines, event_type)
            if a:event_type == 'stderr'
                for l:line in a:lines
                    echom l:line
                endfor
            endif
        endfunction
    """)

    remote_py_code = '\n'.join([
        'import pynvim',
        'nvim = pynvim.attach("stdio")',
        'print("rplugins can write to stdout")',  # tests #377 (#60)
        'nvim.api.command("let g:success = 42")',
    ])
    # see :help jobstart(), *jobstart-options* |msgpack-rpc|
    jobid = vim.funcs.jobstart([
        'python3', '-c', remote_py_code,
    ], {'rpc': True, 'on_stderr': 'OutputHandler'})
    assert jobid > 0
    exitcode = vim.funcs.jobwait([jobid], 500)[0]
    messages = vim.command_output('messages')
    assert exitcode == 0, ("the python process failed, :messages =>\n\n" +
                           messages)

    assert 42 == vim.eval('g:success')
    assert "rplugins can write to stdout" in messages



"""API for working with Nvim tabpages."""

from __future__ import annotations

from typing import Any, TYPE_CHECKING, Tuple

from pynvim.api.common import Remote, RemoteSequence
from pynvim.api.window import Window

if TYPE_CHECKING:
    from pynvim.api.nvim import Nvim


__all__ = ['Tabpage']


class Tabpage(Remote):
    """A remote Nvim tabpage."""

    _api_prefix = "nvim_tabpage_"

    def __init__(self, session: Nvim, code_data: Tuple[int, Any]):
        """Initialize from session and code_data immutable object.

        The `code_data` contains serialization information required for
        msgpack-rpc calls. It must be immutable for Buffer equality to work.
        """
        super(Tabpage, self).__init__(session, code_data)
        self.windows: RemoteSequence[Window] = RemoteSequence(
            self, "nvim_tabpage_list_wins"
        )

    @property
    def window(self) -> Window:
        """Get the `Window` currently focused on the tabpage."""
        return self.request('nvim_tabpage_get_win')

    @property
    def valid(self) -> bool:
        """Return True if the tabpage still exists."""
        return self.request('nvim_tabpage_is_valid')

    @property
    def number(self) -> int:
        """Get the tabpage number."""
        return self.request('nvim_tabpage_get_number')
